//---------------------------------------------------------------------------
// jrdebug_unittesting.h
// functions that support unit testing
//---------------------------------------------------------------------------















//---------------------------------------------------------------------------
// explicit test support functions
void JrDebug::RegisterTestFunction(JrDebugRegisteredTestObj* rtop)
	{
	// register the test function so we can call it later
	if (testobjects!=NULL)
		{
	    lasttestobjp->setNextp(rtop);
		lasttestobjp=rtop;
		}
	else
		{
	    testobjects=rtop;
		lasttestobjp=rtop;
		}
	}

void JrDebug::RunTests(const char *inkeywordstring)
    {
	// iterate through our test object linked list
	jrd_stdstringT keywordstring=jrd_stdstringT(inkeywordstring);
	JrDebugRegisteredTestObj* rtop=testobjects;
	char resultnumbuf[255],timestr[80];
	clock_t startclock = clock();

	// block to run tests
		{
		// create a scoped block object
		jrd_stdstringT announcement = jrd_stdstringT("running tests [")+keywordstring+"]";
		JrDebugBlockObjectData alltestobj(announcement, __FILE__ , __LINE__ ,  __TIMESTAMP__ , __FUNCTION__ , __FUNC__, __func__ , true);

		// loop through applicable tests
		while (rtop!=NULL)
		    {
			if (rtop->doeskeymatch(keywordstring))
			    {
				// set keywords
				curtestobjp=rtop;
				// invoke
				RunTestInternal(rtop,true);
				// advance pointer
				rtop=rtop->getNextp();
				}
			}

		// summarize test results using this test block object data
		int childs=alltestobj.get_childs();
		int childerrors_all=alltestobj.get_childerrors_all();
		int childerrors_block=alltestobj.get_childerrors_block();
		clock_t endclock = clock();
		float elapsed = (float)(endclock - startclock)/(float)CLOCKS_PER_SEC;
		// format nice summary output
		sprintf(timestr,"%.02f",elapsed);
		if (childerrors_all>0)
			sprintf(resultnumbuf,"%d of %d tests with errors, %d err total",childerrors_block,childs,childerrors_all);
		else
			sprintf(resultnumbuf,"all %d tests passed without error",childs);
		}

	// send summary output to stdout after the block ends
	std::cout<< "Testing finished in "<<timestr<< " sec: "<<resultnumbuf<<"."<<std::endl;

	rantests=true;
	}


bool JrDebug::RanTests() {return rantests;};


void JrDebug::ListTests()
	{
	// iterate through our test object linked list
	JrDebugRegisteredTestObj* rtop=testobjects;
	jrd_stdstringT dstring;

	if (!isAnyStdoutEnabled())
		std::cout << std::endl;

	// scoped block announcement
		{
		// create a scoped block object
		jrd_stdstringT announcement = jrd_stdstringT("listing all tests");
		// ATTN: do we want to use jrd outputting or stdout?
		// JrDebugBlockObjectData dummy(announcement, __FILE__ , __LINE__ ,  __TIMESTAMP__ , __FUNCTION__ , __FUNC__, __func__ , true);
		std::cout << announcement<<std::endl;

		// loop through applicable tests
		while (rtop!=NULL)
		    {
			// list function
			dstring=jrd_stdstringT("+")+rtop->getSimpleName()+" ["+rtop->getKeywords()+"]";
			dstring+=jrd_stdstringT(" (")+rtop->getFname()+jrd_stdstringT(" , ")+rtop->getFilename()+"), ";
			dstring+=rtop->getDescription();
			// ATTN: do we want to use jrd outputting or stdout?
			//JrDebug( __FILE__ , __LINE__ ,  __TIMESTAMP__ , __FUNCTION__ , __FUNC__, __func__ , JrdMessage)(JrdInternalInfo,"","",JrDebugDefaultLevel,dstring.c_str());
			// send output to console no matter what
			//if (!isAnyStdoutEnabled())
				std::cout << dstring << std::endl;
			// advance pointer
			rtop=rtop->getNextp();
			}
		}

	rantests=true;
	}

void JrDebug::RunTestInternal(JrDebugRegisteredTestObj* rtop, bool catchexception)
	{
	// run a specific test using rtop pointer
	// start by forcing enable flag
	bool priorenabled = enabled;
	enabled=true;

	// scoped block announcement
		{
		// create a scoped block object
		jrd_stdstringT announcement = jrd_stdstringT(rtop->getDescriptioncharp());
		JrDebugBlockObjectData dummy(announcement, __FILE__ , __LINE__ ,  __TIMESTAMP__ , __FUNCTION__ , __FUNC__, __func__ , true);

		// now invoke the test and catch any exceptions
		if (!catchexception)
			rtop->invoke();
		else
		    {
			// invoke and intercept exceptions
			// Some of these exception types and messages are from boost test library, 12/18/04, execution_monitor.cpp
			//  (C) Copyright Gennadiy Rozental 2001-2003.
			//  (C) Copyright Beman Dawes and Ullrich Koethe 1995-2001.
			//  http://www.boost.org/libs/test for the library home page.
			//  converted using SR "detail::report_error( execution_exception::cpp_exception_error," -> "rtop->AnnounceException("

			try
			    {
				rtop->invoke();
				}

			// build in test exceptions
			catch (JrdTestException &ex)
			    { rtop->AnnounceException((ex.getExceptionString()).c_str()); }

			//  Catch-clause reference arguments are a bit different from function
			catch( char* ex )
				{ rtop->AnnounceException( "C string", ex ); }
			catch( const char* ex )
				{ rtop->AnnounceException( "C string", ex ); }
			catch( jrd_stdstringT const& ex )
				{ rtop->AnnounceException( "jrd_stdstringT", ex.c_str() ); }


			/*
			// I can't get these to reliably compile on enough platforms
			catch( std::bad_alloc const& ex )
				{ rtop->AnnounceException( "std::bad_alloc", ex.what() ); }
			#if !defined(__BORLANDC__) || __BORLANDC__ > 0x0551
				catch( std::bad_cast const& ex )
					{ rtop->AnnounceException( "std::bad_cast", ex.what() ); }
				catch( std::bad_typeid const& ex )
					{ rtop->AnnounceException( "std::bad_typeid", ex.what() ); }
			#else
				catch( std::bad_cast const& ex )
					{ rtop->AnnounceException( "std::bad_cast" ); }
				catch( std::bad_typeid const& ex )
					{ rtop->AnnounceException( "std::bad_typeid" ); }
			#endif

			catch( std::bad_exception const& ex )
				{ rtop->AnnounceException( "std::bad_exception", ex.what() ); }

			// not defined (mingw32) ?
			//			catch( std::domain_error const& ex )
			//				{ rtop->AnnounceException( "std::domain_error", ex.what() ); }
			//			catch( std::invalid_argument const& ex )
			//				{ rtop->AnnounceException( "std::invalid_argument", ex.what() ); }
			//			catch( std::length_error const& ex )
			//				{ rtop->AnnounceException( "std::length_error", ex.what() ); }
			//			catch( std::out_of_range const& ex )
			//				{ rtop->AnnounceException( "std::out_of_range", ex.what() ); }
			//			catch( std::range_error const& ex )
			//				{ rtop->AnnounceException( "std::range_error", ex.what() ); }
			//			catch( std::overflow_error const& ex )
			//				{ rtop->AnnounceException( "std::overflow_error", ex.what() ); }
			//			catch( std::underflow_error const& ex )
			//				{ rtop->AnnounceException( "std::underflow_error", ex.what() ); }
			//			catch( std::logic_error const& ex )
			//				{ rtop->AnnounceException( "std::logic_error", ex.what() ); }
			//			catch( std::runtime_error const& ex )
			//				{ rtop->AnnounceException( "std::runtime_error", ex.what() ); }
			*/

			catch( std::exception const& ex )
				{ rtop->AnnounceException( "std::exception", ex.what() ); }
			catch( ... )
				{ rtop->AnnounceException( "unknown type" ); }
			}
		}

	// return enabled state to what it was before
	enabled=priorenabled;
	}
//---------------------------------------------------------------------------




//---------------------------------------------------------------------------
void JrDebug::RunTestByPointer(JrdFuncpVoidRetT fp)
	{
	// run a specific test using function  pointer
	JrDebugRegisteredTestObj* rtop=testobjects;

	// loop through applicable tests
	while (rtop!=NULL)
		{
		if (rtop->get_fpoint()==fp)
			{
			RunTestInternal(rtop,true);
			break;
			}
		// advance pointer
		rtop=rtop->getNextp();
		}
	if (rtop==NULL)
		{
		// it wasnt found in our registered list, so just invoke it
		fp();
		}
	}
//---------------------------------------------------------------------------































//---------------------------------------------------------------------------
// constructor (for normal scoped construction, used for temp creation of test objects)
JrDebugRegisteredTestObj::JrDebugRegisteredTestObj(const jrd_stdstringT insimplename,const jrd_stdstringT indesc,const jrd_stdstringT inkeywordstring,const jrd_stdstringT infname,JrdFuncpVoidRetT infpoint,const char* infilename, int ilinenum , const char *intimestamp)
{
	nextrtop=NULL;
	simplename=insimplename;
	desc=indesc;
	keywordstring=inkeywordstring;
	fname=infname;
	fpoint=infpoint;
	filename=jrd_stdstringT(infilename);
	timestamp=jrd_stdstringT(intimestamp);
	linenum=ilinenum;
}

// constructor which allocates a new copy and gives it to global debug handler for future management
JrDebugRegisteredTestObj::JrDebugRegisteredTestObj(int dummyval_shouldrealloc,const jrd_stdstringT insimplename,const jrd_stdstringT indesc,const jrd_stdstringT inkeywordstring,const jrd_stdstringT infname,JrdFuncpVoidRetT infpoint,const char* infilename, int ilinenum , const char *intimestamp)
{
	// ATTN: why arent we just registering this object itself, ie JrDebug::RegisterTestFunction(this)?
	//  i think because this is called via macro which creates a temp instance
	JrDebugSingletonInsurer dummy(false);
	JrDebugRegisteredTestObj *objp=new JrDebugRegisteredTestObj(insimplename,indesc,inkeywordstring,infname,infpoint,infilename,ilinenum ,intimestamp);
	JrDebug::RegisterTestFunction(objp);
	// touch unused args to avoid compiler warnings
	JrdUnusedArg(dummyval_shouldrealloc);
}

bool JrDebugRegisteredTestObj::doeskeymatch(const jrd_stdstringT matchstring)
{
	if (matchstring=="all" || matchstring=="*")
		return true;
	if (keywordstring.find(matchstring,0)!=jrd_stdstringT::npos)
		return true;
	return false;
}

// announce exception
void JrDebugRegisteredTestObj::AnnounceException(const char *typestr)
{
	JrDebug( filename.c_str() , linenum , timestamp.c_str() ,  fname.c_str(), "","", JrdMessage) (JrdException,"","",JrDebug::exceptionlevel,"exception thrown during test: %s.",typestr);
}
void JrDebugRegisteredTestObj::AnnounceException(const char *typestr,const char *whatstr)
{
	JrDebug( filename.c_str() , linenum , timestamp.c_str() ,  fname.c_str(), "","", JrdMessage) (JrdException,"","",JrDebug::exceptionlevel,"exception thrown during test: %s [%s].",typestr,whatstr);
}
//---------------------------------------------------------------------------




//---------------------------------------------------------------------------
JrdTestException::JrdTestException(char *inexceptionstring) throw()
	{
	exceptionstring=jrd_stdstringT(inexceptionstring);
	}
JrdTestException::~JrdTestException() throw()
	{
	;
	}
jrd_stdstringT JrdTestException::getExceptionString() throw()
	{
	return exceptionstring;
	}
//---------------------------------------------------------------------------
